# This code numerically checks the convergence of the PURE-CD method
# in the dense matrix case by constructing a Lyapunov function defined
# as the solution of a performance estimation saddle point problem.
#
# We then use this technique to check whether PURE-CD may converge for
# larger step sizes than currently proved.

import numpy as np
import PEPit
from PEPit import operators
from PEPit import functions
from PEPit import primitive_steps
import cvxpy as cp
import dsp

prox = PEPit.primitive_steps.proximal_step
exp2mat = PEPit.tools.expressions_to_matrices.expression_to_matrices

def list_of_leaf_point_names():
    return [PEPit.Point.list_of_leaf_points[i].name for i in range(PEPit.Point.counter)]

def define_lyapunov_from_expressions(V1mV, list_of_expressions_in, list_of_expressions_out):
    assert(len(list_of_expressions_in) == len(list_of_expressions_out))
    q = cp.Variable(len(list_of_expressions_in), 'q')
    for iexpr in range(len(list_of_expressions_in)):
        expr_in_G, expr_in_F, expr_in_C = exp2mat(list_of_expressions_in[iexpr])
        expr_out_G, expr_out_F, expr_out_C = exp2mat(list_of_expressions_out[iexpr])
        V1mV += dsp.inner(q[iexpr], cp.trace(expr_out_G @ G - expr_in_G @ G)
                          + expr_out_F @ F - expr_in_F @ F
                          + expr_out_C - expr_in_C)
    return V1mV, q

GAMMA = np.arange(0.7, 2, 0.02)
gamma = 0.9

solver = 'SCS'
verbose = 1

# min_x max_y sum_i g_i(x_i) + y A_i x_i - h(y), résolu par pure-cd

# Instantiate PEP
problem = PEPit.PEP()

# Declare a smooth convex function
n = 1
L = [1.] * n
h = problem.declare_function(PEPit.functions.ConvexFunction, name='h')
M = [[]] * n
A = [[]] * n
AT = [[]] * n
g = [[]] * n
for i in range(n):
    M[i] = problem.declare_function(PEPit.operators.LinearOperator, L=L[i],
                                    name='A'+str(i))
    A[i] = M[i].gradient
    AT[i] = M[i].T.gradient
    g[i] = problem.declare_function(PEPit.functions.ConvexFunction, name='g'+str(i))

# Declare a partition of the primal-dual space,
#   where the dual space is split into n blocks
partition = problem.declare_block_partition(d=1+n)
# We will have A[i] acting from block i to block n:
#   these constraints will be defined later.

theta = n
tau_ = np.sqrt(gamma * n / theta / np.array(L)**2)
sigma_ = np.max(gamma / np.array(L)**2 / tau_ / n)

# Define a saddle point zs
#   and corresponding Lagrangian value Ls = sum_i g_i(xs[i]) + sum_i ys A[i] xs[i] - f(ys)
free_point_in_lyapunov = True
recenter_saddle_point = False
if recenter_saddle_point:
    zs = PEPit.Point(name='zs', is_leaf=False, decomposition_dict=dict())
    xs = [[]] * n
    for i in range(n):
        xs[i] = PEPit.Point(is_leaf=False, decomposition_dict=dict())
    ys = PEPit.Point(is_leaf=False, decomposition_dict=dict())
else:
    zs = PEPit.Point(name='zs')
    xs = [partition.get_block(zs, i) for i in range(n)]
    ys = partition.get_block(zs, n)

for i in range(n):
    xs[i].name = 'xs'+str(i)
ys.name = 'ys'
gs = [[]] * n
Axs_dev = [[]] * n
for i in range(n):
    Axs_dev[i] = A[i](xs[i])
    Axs_dev[i].name = 'A'+str(i)+'(xs'+str(i)+')'
Axs = np.sum([Axs_dev[i] for i in range(n)])
Axs.name = 'Axs'
ATys = [[]] * n
diff_xs = [[]] * n
for i in range(n):
    gs[i] = PEPit.Expression(name="gs"+str(i))
    ATys[i] = AT[i](ys)
    if free_point_in_lyapunov == False:
        diff_xs[i] = -ATys[i]
    else:
        diff_xs[i] = PEPit.Point(name='diff_xs_i')
    g[i].add_point((xs[i], diff_xs[i], gs[i]))
    ATys[i].name = 'AT'+str(i)+'(ys)'

hs = PEPit.Expression(name="hs")
if free_point_in_lyapunov == False:
    diff_ys = Axs
else:
    diff_ys = PEPit.Point(name='diff_ys')
h.add_point((ys, diff_ys, hs))
Ls = -hs + np.sum([xs[i] * ATys[i] + g[i](xs[i]) for i in range(n)])

counter_for_stationary_point = PEPit.Point.counter

# Then define the point z_{0} of the algorithm
z = problem.set_initial_point('initial point')
x = [partition.get_block(z, i) for i in range(n)]
y = partition.get_block(z, n)
y.name = "y"
Ax_dev = [[]] * n
for i in range(n):
    x[i].name = "x"+str(i)
    Ax_dev[i] = A[i](x[i])
    Ax_dev[i].name = 'A'+str(i)+'(x'+str(i)+')'
Ax = np.sum([Ax_dev[i] for i in range(n)])
Ax.name = 'Ax'

ATy = [[]] * n
for i in range(n):
    ATy[i] = AT[i](y)
    ATy[i].name = "AT"+str(i)+"(y)"

counter_for_initial_point = PEPit.Point.counter

# Define one iteration of pure-cd
yb1, _, _ = prox(y + sigma_ * Ax, h, sigma_)
diff_yb1 = h.list_of_points[-1][1]
diff_yb1.name = 'diff_yb1'
yb1.name = 'yb1'

ATyb1 = [[]] * n
xb1 = [[]] * n
diff_xb1 = [[]] * n
for i in range(n):
    ATyb1[i] = np.sum([AT[i](yb1)])
    ATyb1[i].name = 'ATyb1_'+str(i)
    xb1[i], _, _ = prox(x[i] - tau_[i] * ATyb1[i], g[i], tau_[i])
    xb1[i].name = 'xb1_'+str(i)
    diff_xb1[i] = g[i].list_of_points[-1][1]
    diff_xb1[i].name = 'diff_xb1_'+str(i)

Axb1_dev = [[]] * n
for i in range(n):
    Axb1_dev[i] = A[i](xb1[i])
    Axb1_dev[i].name = 'A'+str(i)+'(xb1_'+str(i)+')'
    
# We now draw a random variable j in {0, ..., n-1} and list
#   all the possible outcomes of the algorithm
y1 = [[]] * n
x1 = [[]] * n
for j in range(n):
    x1[j] = [[]] * n
    for i in range(n):
        if i == j:
            x1[j][j] = xb1[j]
        else:
            x1[j][i] = x[i]
    y1[j] = yb1 + sigma_ * theta * (Axb1_dev[j] - Ax_dev[j])

z1 = [[]] * n
for j in range(n):
    z1[j] = y1[j] + np.sum([x1[j][i] for i in range(n)])
    z1[j].name = 'z1_' + str(j)
counter_for_z1 = PEPit.Point.counter

# Additional quantities for the Lyapunov function

ATy1 = [[]] * n
for j in range(n):
    ATy1[j] = [[]] * n
    for i in range(n):
        ATy1[j][i] = AT[i](y1[j])
        ATy1[j][i].name = 'AT'+str(i)+'(y1_'+str(j)+')'
Ax1_dev = [[]] * n
Ax1 = [[]] * n
for j in range(n):
    Ax1_dev[j] = [[]] * n
    for i in range(n):
        Ax1_dev[j][i] = A[i](x1[j][i])
    Ax1[j] = np.sum([Ax1_dev[j][i] for i in range(n)])

yb2 = [[]] * n
diff_yb2 = [[]] * n
for j in range(n):
    yb2[j], _, _ = prox(y1[j] + sigma_ * Ax1[j], h, sigma_)
    diff_yb2[j] = h.list_of_points[-1][1]
    diff_yb2[j].name = 'diff_yb2_'+str(j)
    yb2[j].name = 'yb2_'+str(j)

ATyb2 = [[]] * n
xb2 = [[]] * n
diff_xb2 = [[]] * n
for j in range(n):
    ATyb2[j] = [[]] * n
    xb2[j] = [[]] * n
    diff_xb2[j] = [[]] * n
    for i in range(n):
        ATyb2[j][i] = np.sum([AT[i](yb2[j])])
        ATyb2[j][i].name = 'ATyb2_'+str(j)+'_'+str(i)
        xb2[j][i], _, _ = prox(x1[j][i] - tau_[i] * ATyb2[j][i], g[i], tau_[i])
        xb2[j][i].name = 'xb2_'+str(j)+'_'+str(i)
        diff_xb2[j][i] = g[i].list_of_points[-1][1]
        diff_xb2[j][i].name = 'diff_xb2_'+str(j)+'_'+str(i)
Axb2_dev = [[]] * n
for j in range(n):
    Axb2_dev[j] = [[]] * n
    for i in range(n):
        Axb2_dev[j][i] = A[i](xb2[j][i])
        Axb2_dev[j][i].name = 'A'+str(i)+'(xb2_'+str(j)+'_'+str(i)+')'

# We now draw another random variable j2 in {0, ..., n-1} and list
#   all the possible outcomes of the algorithm
y2 = [[]] * n
x2 = [[]] * n
for j in range(n):
    x2[j] = [[]] * n
    y2[j] = [[]] * n
    for j2 in range(n):
        x2[j][j2] = [[]] * n
        for i in range(n):
            if i == j2:
                x2[j][j2][j2] = xb2[j][j2]
            else:
                x2[j][j2][i] = x1[j][i]
        y2[j][j2] = yb2[j] + sigma_ * theta * (Axb2_dev[j][j2] - Ax1_dev[j][j2])

z2 = [[]] * n
for j in range(n):
    z2[j] = [[]] * n
    for j2 in range(n):
        z2[j][j2] = y2[j][j2] + np.sum([x2[j][j2][i] for i in range(n)])
        z2[j][j2].name = 'z2_' + str(j) + '_' + str(j2)
counter_for_z2 = PEPit.Point.counter


# xb, yb, x_next, y_next, etc should remain in their respective primal or dual spaces

list_of_primal_points = [xs, x, ATy, ATys, diff_xs, ATyb1, xb1, diff_xb1] + ATy1 + xb2 + ATyb2
list_of_dual_points = [ys, diff_ys, y, yb1, diff_yb1] + Axs_dev + Ax_dev + Axb1_dev + yb2
for j in range(n):
    list_of_dual_points += Axb2_dev[j]

for y_ in list_of_dual_points:
    for x_ in list_of_primal_points:
        for i in range(n):
            problem.add_constraint(x_[i] * y_ == 0, 'partition')
for x1_ in list_of_primal_points:
    for x2_ in list_of_primal_points:
        for j1 in range(n):
            for j2 in range(j1):
                problem.add_constraint(x1_[j1] * x2_[j2] == 0, 'partition')


Ddyb1 = h(yb1) - h(ys) - diff_ys * (yb1 - ys)
EDdyb2 = 1/n * sum([h(yb2[j]) - h(ys) - diff_ys * (yb2[j] - ys) for j in range(n)])
Dpx = sum([g[i](x[i]) - g[i](xs[i]) - diff_xs[i] * (x[i] - xs[i]) for i in range(n)])
EDpx1 = 1/n * sum([
    sum([g[i](x1[j][i]) - g[i](xs[i]) - diff_xs[i] * (x1[j][i] - xs[i])
         for i in range(n)]) for j in range(n)])

final_counter = PEPit.Point.counter

def norm_V(x1, y1, x2, y2, tau, sigma):
    return 0.5 * sum([(x1[i]-x2[i])**2/tau[i] for i in range(n)]) \
        + (y1 - y2)**2 / sigma

problem.set_initial_condition((1-1/n)*Dpx + norm_V(x, y, xs, ys, tau_, sigma_) <= 1, 'initial condition')

# Set an arbitrary performance metric
problem.set_performance_metric(0 * Ls)  # norm_V(x1[0], ATy1[0], y1[0], xs, ATys, ys, tau_, sigma_))

# Solve the silly PEP to recover a cvxpy object coding the constraints
print("Constructing CVXPY objects from PEPit.")
pepit_verbose = 0
problem.solve(wrapper="cvxpy", solver=solver, verbose=pepit_verbose)
cvxpy_prob = problem.wrapper.prob

# Constructing next iterate matrix Sigma
Sigma = [[]] * n
for j in range(n):
    Sigma[j] = exp2mat(
        sum([PEPit.Point.list_of_leaf_points[a]**2
                     for a in range(counter_for_stationary_point)]))[0]
    Sigma[j] += exp2mat(z*z1[j])[0]
    for i in range(n):
        Sigma[j] += exp2mat(x[i]*x1[j][i])[0]
    Sigma[j] += exp2mat(Ax_dev[j]*Axb1_dev[j])[0]
    for i in range(n):
        Sigma[j] += exp2mat(ATy[i]*ATy1[j][i])[0]
    Sigma[j] += exp2mat(diff_yb1*diff_yb2[j])[0]
    for i in range(n):
        Sigma[j] += exp2mat(ATyb1[i]*ATyb2[j][i])[0]
        Sigma[j] += exp2mat(diff_xb1[i]*diff_xb2[j][i])[0]
        Sigma[j] += exp2mat(Axb1_dev[i]*Axb2_dev[j][i])[0]
    Sigma[j] = np.triu(Sigma[j]) * 2 - np.diag(np.diag(Sigma[j]))

# Declaring cvxpy variables for the Lyapunov function and the residual function

co = counter_for_z1
Q = cp.Variable((co, co), 'Q', PSD=True)

F = cvxpy_prob.variables()[0]
G = cvxpy_prob.variables()[1]

Pzzs, _, _ = exp2mat(norm_V(x, y, xs, ys, tau_, sigma_))
Pz1_0_zs, _, _ = exp2mat(norm_V(x1[0], y1[0], xs, ys, tau_, sigma_))
Pzzb, _, _ = exp2mat(norm_V(x, y, xb1, yb1, tau_, sigma_))

Qplus = 0*Pzzs[:co, :co] + Q

V1mV = dsp.inner(Qplus.flatten(), 1./n * sum([(Sigma[j] @ G @ Sigma[j].T - G)[:co, :co].flatten() for j in range(n)]))
Ex1mx = np.sum([np.sum([(x1[j][i] - x[i])**2 for j in range(n)]) for i in range(n)]) / n
Ex2mx1 = np.sum([np.sum([np.sum([(x2[j][j2][i] - x1[j][i])**2 for j in range(n)]) for i in range(n)]) for j2 in range(n)]) / n**2
Ey1my = np.sum([(y1[j] - y)**2 for j in range(n)]) / n
Ey2my1 = np.sum([np.sum([(y2[j][j2] - y1[j])**2 for j in range(n)]) for j2 in range(n)]) / n**2

V1mV, q = define_lyapunov_from_expressions(V1mV, [Dpx, Ddyb1, Ex1mx, Ey1my], [EDpx1, EDdyb2, Ex2mx1, Ey2my1])

# Set constraints on Q and q
constraints = [q >= 0]
constraints += [Q << 10000 * np.eye(co), q <= 10000]

R = cp.trace(Pzzb @ G)

obj = dsp.MinimizeMaximize(V1mV + R) # + 0.001 * (cp.trace(Q) + cp.sum(q)))

dsp_problem = dsp.SaddlePointProblem(obj, constraints + cvxpy_prob.constraints, minimization_vars=[Q, q], maximization_vars=cvxpy_prob.variables())

# solve the problem
print("Solving saddle point problem using DSP-CVXPY.")

try:
    dsp_problem.solve(eps=1e-4, verbose=True, max_iters=1000000)
except Exception as err:
    print("---------- error in DSP -----------")
    print(err)
    print("-----------------------------------")

problem._eval_points_and_function_values(cvxpy_prob.variables()[0].value,
                                         cvxpy_prob.variables()[1].value)
problem.wrapper.assign_dual_values()

print("Saddle point problem value is " + str((V1mV + R).value) + " and should be 0.")

np.set_printoptions(linewidth=200)


def simple_example(gamma=0.99, K=100, n=2):
    # Pour n > 1, on peut aller plus loin que gamma=1, et même jusqu'à gamma < 2 pour n grand...
    
    # the case min_{x1, x2} max_y y(x1+x2)
    x = np.ones(n)
    y = 1
    A = np.ones(n)
    sigma = np.sqrt(gamma / n)
    tau = sigma
    Z = np.zeros(K)
    for k in range(K):
        x_ = x.copy()
        yb = y + sigma * A @ x_
        xb = x_ - tau * A.T * yb
        i = np.random.randint(n)
        x[i] = xb[i]
        y = yb + n * sigma * A @ (x - x_)
        Z[k] = abs(sum(x)) + abs(y)
    return x, y, Z


def separate_blocks(points_grouped_by_block):
    eps = 1e-5
    d = len(points_grouped_by_block)
    points = points_grouped_by_block
    new_points = [[]] * d
    for j in range(d):
        M = np.vstack([points[j][i].eval() for i in range(len(points[j]))])
        U, S, V = np.linalg.svd(M)
        rank = np.sum(S > eps)
        new_points[j] = U @ np.diag(S)[:, :rank]

    return new_points

list_of_primal_points_transposed = [ [list_of_primal_points[i][j] for i in range(len(list_of_primal_points))] for j in range(n)]
new_points = separate_blocks([list_of_dual_points] + list_of_primal_points_transposed)
